# Copyright © 2023 Neven Sajko

module Minimax

import
  MathOptInterface,
  Polynomials,
  ..NumericalErrorTypes,
  ..PolynomialPassingThroughIntervals,
  ..Perturbed,
  ..IntervalHelpers,
  ..ToSparsePolynomial

const MOI = MathOptInterface
const NE = NumericalErrorTypes
const PPTI = PolynomialPassingThroughIntervals
const ItvH = IntervalHelpers

numerical_error_unchecked(::NE.AbsoluteError, accurate, approximate) =
  abs(accurate - approximate)

function numerical_error_unchecked(::NE.RelativeError, accurate, approximate)
  (accurate == approximate) && return zero(approximate)
  abs(1 - approximate/accurate)
end

function numerical_error(error_type::NE.NumericalErrorType, accurate, approximate)
  isfinite(accurate) || error("accurate not finite")
  isfinite(approximate) || error("approximate not finite")
  numerical_error_unchecked(error_type, accurate, approximate)
end

check_val_tiny(::NE.AbsoluteError, ::F, ::F) where {F <: Real} = nothing

function check_val_tiny(::NE.RelativeError, val::F, maximal_tiny::F) where {F <: Real}
  (abs(val) < maximal_tiny) && error("value too tiny")
  nothing
end

function check_interval_endpoint(
  endpoint::F,
  fun::Fun,
  error_type::NE.NumericalErrorType,
  maximal_tiny::F,
) where {F <: Real, Fun <: Function}
  val = fun(endpoint)
  isfinite(val) || error("interval endpoint not finite")
  check_val_tiny(error_type, val, maximal_tiny)
  nothing
end

check_interval_endpoint(
  fun::Fun,
  error_type::NE.NumericalErrorType,
  maximal_tiny::Real,
) where {Fun <: Function} =
  let t = maximal_tiny
    ep -> check_interval_endpoint(ep, fun, error_type, t)
  end

max_err_on_grid(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  poly::Polynomials.AbstractPolynomial{F},
  grid::Vector{F},
) where {
  Fun <: Function,
  F <: AbstractFloat,
} =
  mapreduce(
    let p = poly
      x -> numerical_error(error_type, fun(x), evalpoly(x, p))
    end,
    max,
    grid,
    init = zero(F),
  )

struct WorstNontinyPointMapreduceHelper{F <: Real}
  input::F

  # A negative value is special
  error::F
end

WorstNontinyPointMapreduceHelper{F}() where {F <: Real} =
  WorstNontinyPointMapreduceHelper{F}(zero(F), -one(F))

function WorstNontinyPointMapreduceHelper{F}(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  x::F,
  maximal_tiny::F,
) where {
  Fun <: Function,
  F <: Real,
}
  accurate = fun(x)
  (abs(accurate) < maximal_tiny) && return WorstNontinyPointMapreduceHelper{F}()
  approximate = evalpoly(x, pol)
  WorstNontinyPointMapreduceHelper{F}(
    x,
    numerical_error(error_type, accurate, approximate),
  )
end

worst_nontiny_point_reduce_bin_op(
  l::T,
  r::T,
) where {T <: WorstNontinyPointMapreduceHelper} =
  (l.error < r.error) ? r : l

worst_nontiny_point(
  error_type::NE.NumericalErrorType,
  fun::Fun,
  pol::Polynomials.AbstractPolynomial{F},
  grid::Vector{F},
  maximal_tiny::F,
) where {
  Fun <: Function,
  F <: Real,
} =
  mapreduce(
    let mt = maximal_tiny, p = pol
      (x::F) -> WorstNontinyPointMapreduceHelper{F}(error_type, fun, p, x, mt)
    end,
    worst_nontiny_point_reduce_bin_op,
    grid,
    init = WorstNontinyPointMapreduceHelper{F}(),
  )

"""
minimax_polynomial
  * Takes:
    * A function that returns a new MOI.AbstractOptimizer
      * A new MathOptInterface LP optimizer each time it's called
      * Only one will be used at a time
      * Tested with Tulip
    * A function to be approximated as input
      * Only univariate real functions are supported
    * An `NTuple{2, F}` that represents the interval over which
      the given function is to be approximated
      * `F` subtypes AbstractFloat
      * The first value in the tuple is the lower bound of the interval, and
        the second value is the upper bound
      * None of the two bounds of the interval may be a root of the given
        function when the error type is relative error. If this is a problem,
        the user should try slightly shrinking or expanding the interval over
        the root.
    * The indices denoting the monomials that the sought-after polynomial is
      supposed to have
      * For example, use `0:3` to look for any polynomial of degree three, or
        `0:2:8` for any even polynomial of degree eight
      * Note: the user should use even polynomials for approximating even
        functions and odd polynomials for approximating odd functions
    * A hard limit on the number of iterations for the algorithm, to prevent it
      from entering an infinite loop or just running too long.
    * The type of the error that we're minimizing, absolute or relative
      * Optional argument
    * The upper bound on allowed maximum error
      * Optional argument
    * Initial LP grid
      * Optional argument
      * This gets iteratively improved, at each step of the algorithm
      * When one is minimizing the absolute error, it might make sense to just
        pass the Chebyshev nodes here
    * Options for the LP formulation
      * Optional argument
    * Factor for perturbing the grid (1)
      * Keyword argument
    * Segment grid point count (1)
      * Keyword argument
      * The higher this value is, the more accurate the algorithm should be,
        producing a polynomial closer to the accurate minimax. If the value is
        too low, high-error regions might go unnoticed by the algorithm,
        especially if the function targeted for the approximation isn't very
        smooth.
    * Factor for perturbing the grid (2)
      * Keyword argument
    * Segment grid point count (2)
      * Keyword argument
    * Exit condition
      * Keyword argument
    * Maximal value considered to be too "tiny" for LP
      * Keyword argument
  * Returns a named tuple:
    * A vector, `coefficients`:
      * If nonempty, it's the coefficients of the polynomial in the standard
        basis
      * If empty, the last LP solve failed
        * This might mean that more monomials are needed
        * This might mean the arithmetic precision is too low and should be
          increased to enable the algorithm to complete successfully
    * The final LP grid, `grid`:
      * May be used for refining the result further, e.g., by running again
        with an increased segment grid point count value
      * May be used for finishing where the algorithm left off after a failure
        caused by a lack of precision (increase the precision first, though)
    * The most recently used LP optimizer, `lp`
    * A vector containing the max error after each step, `max_errors`

Usage example:
```
# We'll be using BigFloat, so increase it's precision a bit
setprecision(BigFloat, 6 * 2^7)

import FindMinimaxPolynomial, Tulip, MathOptInterface

# Shorthands
const FMP = FindMinimaxPolynomial
const MMX = FMP.Minimax
const PPTI = FMP.PolynomialPassingThroughIntervals
const NE = FMP.NumericalErrorTypes
const to_poly = FMP.ToSparsePolynomial.to_sparse_polynomial
const mmx = MMX.minimax_polynomial
const error_type_relative = NE.RelativeError()
const MOI = MathOptInterface

function make_lp()
  lp = Tulip.Optimizer{BigFloat}()

  # Drastically increase iteration limit
  MOI.set(lp, MOI.RawOptimizerAttribute("IPM_IterationsLimit"), 2000)

  # Disable presolve
  MOI.set(lp, MOI.RawOptimizerAttribute("Presolve_Level"), 0)

  lp
end

# The monomials for an odd polynomial with degree 7
monomials_4_odd = 1:2:7

# The interval on which we'll approximate the sine
itv_sin = (big"0.1", big"0.2");

# Approximate the sine from 0.1 to 0.2 with such a polynomial
res_sin = mmx(
  make_lp,
  sin,
  itv_sin,
  monomials_4_odd,
);

# Convert the coefficients into a polynomial for evaluation, plotting, etc.
poly_sin = to_poly(res_sin.coefficients, monomials_4_odd)

# The monomials for an even polynomial with degree 6
monomials_4_even = 0:2:6

# The interval on which we'll approximate the cosine
itv_cos = (big"1.0", big"2.0");

# Roots don't bother us! (The cosine has a root at π/2, which is on the chosen
# interval.)
res_cos = mmx(
  make_lp,
  cos,
  itv_cos,
  monomials_4_even,
);

poly_cos = to_poly(res_cos.coefficients, monomials_4_even)

# Let's plot our approximation errors!

using Gadfly

err_sin(x) =
  let x = BigFloat(x)
    abs(1 - poly_sin(x)/sin(x))
  end

err_sin_max(x) = last(res_sin.max_errors)

plot([err_sin, err_sin_max], Float64.(itv_sin)...)

err_cos(x) =
  let x = BigFloat(x)
    abs(1 - poly_cos(x)/cos(x))
  end

err_cos_max(x) = last(res_cos.max_errors)

plot([err_cos, err_cos_max], Float64.(itv_cos)...)
```
"""
function minimax_polynomial(
  make_lp::LP,
  fun::Fun,
  interval::NTuple{2, F},
  monomials::AbstractVector{<:Integer},
  iteration_limit::Int = 200,
  error_type::NE.NumericalErrorType = NE.RelativeError(),
  max_error_upper_bound::F = typemax(F),
  ret_grid::Vector{F} =
    Perturbed.perturbed_grid_on_interval((length(monomials) + 1) * 1, interval, 1//8),
  ::LPOption = PPTI.ReplacedVariablesOption{:determinants}();
  worst_segments_perturb_factor::Real = 1//8,
  worst_segments_grid_point_count::Int = 2^16,
  within_segment_perturb_factor::Real = 1//8,
  within_segment_grid_point_count::Int = 2^18,
  exit_condition::F = eps(F)^(1//8),
  maximal_tiny::F = floatmin(F)^(1//32),
  worst_segment_count::Int = 3,
) where {
  LP <: Function,
  Fun <: Function,
  F <: AbstractFloat,
  LPOption <: PPTI.FeasibleOption,
}
  # * Note: when calculating relative error, first check whether the arguments
  #   are equal and return zero if so. Otherwise there would be problems when
  #   both arguments are zero.
  #
  # * Handling singularities:
  #   * The function could evaluate to a value that is not finite, which causes
  #     an error to be thrown
  #   * If the function evaluates to zero or to a small value, the error could
  #     evaluate to infinity or to a very large number, which we check by
  #     comparing the function value with the maximal "tiny" value
  #     * In the case of absolute error, this will not happen
  #     * In the case of relative error:
  #       * If the point is one of the two edges of the interval, this is an
  #         error
  #       * If the point is meant for computing relative error, just compute
  #         the error, as long as it's not NaN it's a valid error. However, now
  #         it's very likely that we'll be putting an LP point somewhere around
  #         here...
  #       * If the point is meant for LP, we need to find another point close
  #         by, because tiny intervals could cause numerical issues for LP.

  foreach(check_interval_endpoint(fun, error_type, maximal_tiny), interval)

  ret_coefs = F[]
  ret_lp = make_lp()
  ret_errs = F[]

  # Something like this:
  #
  # 1. Do LP with the grid. Exit if there's no solution.
  # 2. Test the resulting polynomial's error with a denser grid
  # 3. Pick a segment between some LP grid points such that the maximal error
  #    on the segment is greater than the max error on LP grid
  # 4. Add a point for LP within the chosen segment
  # 5. Go back to step 1
  iteration_count = 1
  while true
    func_vals = map(fun, ret_grid)

    lp_intervals = ItvH.intervals_given_max_err(
      func_vals,
      error_type,
      max_error_upper_bound,
    )

    # LP
    ret_coefs = PPTI.polynomial_passing_through_intervals!(
      ret_lp,
      monomials,
      ret_grid,
      lp_intervals,
      F,
      PPTI.MinimaxOption{LPOption}(func_vals, error_type, max_error_upper_bound),
    )

    # Exit if the LP failed
    isempty(ret_coefs) && break

    polynomial = ToSparsePolynomial.to_sparse_polynomial(ret_coefs, monomials)

    max_lp_err = max_err_on_grid(error_type, fun, polynomial, ret_grid)

    max_max_err = -one(F)

    # 1. Int: index into ret_grid representing the segment
    #
    # 2. F: maximal found error on the segment
    worst_segments = Tuple{Int, F}[]

    # At the same time:
    #   * Look for max max error
    #   * Look for worst_segment_count worst segments
    for segment_index ∈ eachindex(ret_grid)[begin:(end - 1)]
      segment = map(
        let g = ret_grid
          i -> ret_grid[i]
        end,
        (segment_index, segment_index + 1))
      worst_segments_grid = Perturbed.perturbed_grid_on_interval(
        worst_segments_grid_point_count,
        segment,
        worst_segments_perturb_factor,
      )
      pop!(worst_segments_grid)
      popfirst!(worst_segments_grid)  # Prevent duplicates in LP grid
      segment_max_err = max_err_on_grid(error_type, fun, polynomial, worst_segments_grid)

      # Update max_max_err as necessary
      (max_max_err < segment_max_err) && (max_max_err = segment_max_err)

      # Update worst_segments as necessary
      (max_lp_err ≤ segment_max_err) &&
        push!(worst_segments, (segment_index, segment_max_err))
      sort!(
        worst_segments,
        by = last,
        rev = true,
      )
      (worst_segment_count < length(worst_segments)) && pop!(worst_segments)
    end

    push!(ret_errs, max_max_err)

    all(isfinite, (max_max_err, max_lp_err)) &&
    (
      (max_max_err ≤ max_lp_err) ||
      (
        numerical_error(
          NE.RelativeError(),
          max_max_err,
          max_lp_err,
        ) < exit_condition
      )
    ) &&
      # Success!
      break

    (iteration_limit ≤ iteration_count) && break

    # Pick a segment from worst_segments randomly
    (chosen_segment_index, unused_err) = worst_segments[rand(eachindex(worst_segments))]

    # Take a point from the chosen segment
    #   1. Take a perturbed equidistant grid on the segment
    #   2. Evaluate the error on each point of the grid
    #   3. Among the non-tiny points, take the one with maximal error. If all
    #      points are tiny, error.
    within_segment_grid = Perturbed.perturbed_grid_on_interval(
      within_segment_grid_point_count,
      (
        ret_grid[chosen_segment_index],
        ret_grid[chosen_segment_index + 1],
      ),
      within_segment_perturb_factor,
    )
    pop!(within_segment_grid)
    popfirst!(within_segment_grid)  # Prevent duplicates in LP grid
    new_point = worst_nontiny_point(
      error_type,
      fun,
      polynomial,
      within_segment_grid,
      maximal_tiny,
    ).input

    # Add the chosen point to the LP grid
    insert!(ret_grid, chosen_segment_index + 1, new_point)

    # TODO: remove this after testing
    issorted(ret_grid) ||
      error("debugging: LP grid not sorted, it_cnt:$iteration_count")
    allunique(ret_grid) ||
      error(
        "debugging: LP grid has repeated identical elements\n" *
        "it_cnt:$iteration_count")

    ret_lp = make_lp()
    iteration_count += 1
  end

  (
    coefficients = ret_coefs,
    grid = ret_grid,
    lp = ret_lp,
    max_errors = ret_errs,
  )
end

end
